{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {
    "tags": [
     "remove_cell"
    ]
   },
   "outputs": [],
   "source": [
    "from IPython.display import display, IFrame\n",
    "def example(url): display(IFrame(url, width=\"100%\", height=600))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Теоретичні відомості\n",
    "## Масиви типу `char`, або символьні масиви\n",
    "\n",
    "Мова С не містить стандартного типу даних *рядок*. Замість цього вона підтримує масиви символів, які завершуються нуль-символом `'\\0'` (символ з ASCII-кодом 0). Функції для роботи з такими масивами описано в заголовковому файлі `string.h`. Вони дають змогу досягти високої ефективності, але небезпечні у використанні, оскільки вихід за межі рядка не перевіряється.\n",
    "\n",
    "Розташування нуль-символа визначає фактичну довжину рядка. Наприклад, довжина рядка `char text[] = \"Моя програма!\"` дорівнює 25 байтів."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "tags": [
     "popout"
    ]
   },
   "source": [
    "Тут необхідно згадати, що кирилічні символи кодуються багатобайтовими послідовностями. У випадку UTF-8 це 2 байти на символ, звідси $2 * 11 + 3 = 25$, де три однобайтових символи це пробіл, знак оклику та прикінцевий нуль-символ.\n",
    "\n",
    "Стандартні функції роботи з рядками зі `string.h` очікують, що символи закодовані у ASCII-коді. Для маніпулювання UTF-8 текстом необхідно використовувати сторонні бібліотеки, наприклад [`utf8.h`](https://www.cprogramming.com/tutorial/unicode.html).\n",
    "\n",
    "Часто, навіть якщо ви правильно маніпулюєте багатобайтовими кодуваннями, вони можуть відображатися у консолі некоректно — як `?` чи `□` — через настройки кодової стрінки самого терміналу.\n",
    "\n",
    "На Windows це можна виправити виконавши команду [`chcp 65001`](https://superuser.com/questions/269818/change-default-code-page-of-windows-console-to-utf-8). Інструкції для [MacOS](https://stackoverflow.com/a/4606636/5679064). Інструкції для різних дистрибутивів та емуляторів терміналів у Linux можуть відрізнятися, наприклад, для [Ububtu](https://askubuntu.com/questions/87227/switch-encoding-of-terminal-with-a-command)."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Символьна константа — це символ (один), оточений одинарними лапками, наприклад, `'Х'`. Значенням символьної константи є чисельне значення цього символу в машинному поданні набору символів. Усі символи впорядковано відповідно до кодів ASCII. При цьому порядковий номер символу називають *кодом* (наприклад, код латинського символу `'А'` дорівнює 65; код символу `'3'` дорівнює 51). Для символьних даних не визначено арифметичних операцій, але їх можна порівнювати за своїми кодами, використовувати в читанні, друку, операторах присвоювання.\n",
    "\n",
    "У програмі рядки можна визначати таким чином:\n",
    "- як рядкові константи;\n",
    "- як масиви символів;\n",
    "- через указівник на символьний тип;\n",
    "- як масиви рядків.\n",
    "\n",
    "Для зберігання рядка обов’зково має бути передбачено виділення пам’яті.\n",
    "\n",
    "Будь-яка послідовність символів, оточена в подвійні лапки `\"\"`, розглядається як *рядкова константа*. Для коректного виведення будь-який рядок повинен закінчуватися нуль-символом `'\\0'`. Під час оголошення рядкової константи нуль-символ додається до неї автоматично. Під зберігання рядка виділяють елементи оперативної пам’яті, що йдуть послідовно. Таким чином, рядок є масивом символів. Для зберігання коду кожного символу рядка відводиться 1 байт.\n",
    "\n",
    "Для розміщення в рядковій константі деяких службових символів використовують символьні комбінації. Зокрема, якщо потрібно включити в рядок символ подвійної лапки, йому повинен передувати символ `'\\'` (зворотний слеш). Рядкові константи розміщують у статичній пам’яті. Початкову адресу послідовності символів у подвійних лапках трактують як адресу рядка.\n",
    "\n",
    "Під час визначення *масиву символів* потрібно повідомити компілятору розмір пам’яті, який треба відвести під нього, наприклад:\n",
    "```c\n",
    "char m[82];\n",
    "```\n",
    "Компілятор також може самостійно визначити розмір масиву символів, якщо ініціалізацію масиву задано під час оголошення за допомогою рядкової константи."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```{margin}\n",
    "Запустити цей приклад можна за [посиланням](https://repl.it/@GasperPaul/Lab3-Example-31).\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Приклад 3.1\n",
    "\n",
    "Розгляньмо декларацію та ініціалізацію рядка:\n",
    "```c\n",
    "#include <stdio.h>\n",
    "\n",
    "char cString1[] = \"A new string is:\";\n",
    "char cString2[] = { 'A', ' ', 'n', 'e', 'w', ' ', 's', 't', 'r', \n",
    "                    'i', 'n', 'g', ' ', 'i', 's', ':', '\\0' };\n",
    "\n",
    "int main(void)\n",
    "{\n",
    "    printf(\"%s\\n\", cString1);\n",
    "    scanf(\"%s\", cString1);\n",
    "    printf(\"\\n%s\\n%s\", cString1, cString2);\n",
    "}\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Варто звернути увагу на виклики функції `printf()`. Параметр форматування `%s` змінює її синтаксис: тепер аргументом має бути вказівник на рядок. Функція `scanf()` завершує уведення після натискання пробілу або клавіші `Enter`.\n",
    "\n",
    "Рядки `cString1` і `cString2` абсолютно ідентичні під час ініціалізації. Символьний масив `cString1` складається з 17 елементів, хоча друкованих символів у ньому 16. Ця невідповідність пояснюється керуючим символом, що позначає кінець рядка. У випадку формування символьного масиву в процесі виконання програми потрібно поклопотатися про коректне закінчення рядка, оскільки бібліотекові функції використовують цю ознаку.\n",
    "\n",
    "Для задавання рядка можна використовувати вказівник на символьний тип. Іноді в програмах виникає потреба опису масиву символьних рядків. У цьому випадку можна використовувати індекс рядка для доступу до декількох різних рядків.\n",
    "\n",
    "Більшість операцій мови С, що мають справу з рядками, працює зі вказівниками. Для розміщення в оперативній пам’яті рядка символів потрібно:\n",
    "- виділити під масив блок оперативної пам’яті;\n",
    "- ініціалізувати рядок.\n",
    "\n",
    "Для виділення пам’яті під зберігання рядка можна використовуватифункції динамічного виділення пам’яті. При цьому потрібно враховувати розмір рядка:\n",
    "```c\n",
    "char* name;\n",
    "name = (char*) malloc(10);\n",
    "scanf(\"%9s\", name);\n",
    "```\n",
    "У вищенаведеному фрагменті для уведення рядка використано функцію `scanf()`, при цьому уведений рядок не може перевищувати 9 символів. Останній символ рядка буде `'\\0'`.\n",
    "\n",
    "## Функції уведення та виведення рядків\n",
    "\n",
    "Для уведення рядка можна використовувати функцію `scanf()`. Проте, функцію `scanf()` призначено радше для отримання не рядка, а окремого слова. Якщо застосовувати формат `\"%s\"` для уведення, рядок уводиться до (але не включаючи) наступного порожнього символу, яким може бути пробіл, табуляція або символ нового рядка.\n",
    "\n",
    "Для уведення рядка, включаючи пробіли, використовують таку функцію:\n",
    "```c\n",
    "char* gets(char*);\n",
    "```\n",
    "Указівник на рядок, у який здійснюється уведення, передають як аргумент функції. Функція просить користувача увести рядок, який вона поміщає в масив. Уведення завершується, коли користувач натисне клавішу `Enter`.\n",
    "\n",
    "*Зауваження 1*. Функція `gets` читає все, що набирає користувач, доти, доки він не натисне клавішу `Enter`. Код клавіші `Enter` не розміщується в рядку. У кінець рядка додається символ `'\\0'`.\n",
    "\n",
    "*Зауваження 2*. Якщо під час уведення рядка кількість знаків перевищить його розмір, установлений під час його опису, то це швидше за все призведе до збою роботи програми, оскільки зайві знаки буде записано в пам’ять «чужої» області, не зарезервованої під рядок.\n",
    "\n",
    "Для виведення рядків можна використати розглянуту раніше функцію `printf(\"%s\", str)`, або її скорочений формат: `printf(str)`, де `str` — указівник на рядок.\n",
    "\n",
    "Для виведення рядків також можна використовувати таку функцію:\n",
    "```c\n",
    "int puts(char* s);\n",
    "```\n",
    "Функція друкує рядок `s` і переводить курсор на новий рядок (на відміну від `printf()`). Функцію `puts()` також можна використовувати для виведення рядкових констант, оточених лапками.\n",
    "\n",
    "## Функції уведення та виведення символів\n",
    "\n",
    "Для уведення символів можна використовувати таку функцію:\n",
    "```c\n",
    "char getchar();\n",
    "```\n",
    "Функція повертає значення символу, уведеного з клавіатури. Зазначену функцію використовують для затримки вікна консолі після виконання програми до натискання клавіші.\n",
    "\n",
    "Для виведення символів можна використовувати таку функцію:\n",
    "```c\n",
    "char putchar(char c);\n",
    "```\n",
    "Функція повертає значення символу, що виводиться, і виводить на екран символ, переданий у якості аргумента."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```{margin}\n",
    "Запустити цей код можна за [посиланням](https://repl.it/@GasperPaul/Lab3-Example-32).\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Приклад 3.2\n",
    "\n",
    "У наступній програмі реалізується підрахунок кількості символів в уведеному рядку.\n",
    "\n",
    "```c\n",
    "#include <stdio.h>\n",
    "int main()\n",
    "{\n",
    "    char str[80], sym;\n",
    "    int count;\n",
    "\n",
    "    printf(\"Уведіть рядок: \");\n",
    "    gets(str);\n",
    "    printf(\"Уведіть символ: \");\n",
    "    sym = getchar();\n",
    "\n",
    "    count = 0;\n",
    "    for (int i = 0; str[i] != '\\0'; i++) {\n",
    "        if (str[i] ==  sym)\n",
    "            count++;\n",
    "    }\n",
    "    printf(\"У рядку\\n\");\n",
    "    puts(str);    /* виведення рядка */\n",
    "    printf(\"символ \");\n",
    "    putchar(sym); /* виведення символу */\n",
    "    printf(\" зустрічається %d разів\\n\", count);\n",
    "    getchar();\n",
    "\n",
    "    return 0;\n",
    "}\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Основні функції бібліотеки `string.h` наведено в таблиці 3.1.\n",
    "\n",
    "Таблиця 3.1 – Функції бібліотеки `string.h`\n",
    "\n",
    "| Функція | Опис |\n",
    "| ----    | ---- |\n",
    "| `char* strcat(char* s1, char* s2)` | приєднує `s2` до `s1`, повертає `s1` |\n",
    "| `char* strncat(char* s1, char* s2, int n)` | приєднує не більше за `n` символів `s2` до `s1`, завершує рядок символом `'\\0'`, повертає `s1` |\n",
    "| `char* strсpy(char* s1, char* s2)` | копіює `s2` в `s1`, включаючи `'\\0'`, повертає `s1` |\n",
    "| `char* strncpy(char* s1, char* s2, int n)` | копіює не більше за `n` символів `s2` в `s1`, повертає `s1` |\n",
    "| `intstrcmp(char* s1, char* s2)` | порівнює `s1` і `s2`, повертає значення 0, якщо рядки еквівалентні |\n",
    "| `intstrncmp(char* s1, char* s2, int n)` | порівнює початкові `n` символів `s1` і `s2`, повертає значення 0, якщо початкові `n` символів рядків еквівалентні |\n",
    "| `intstrlen(char* s)` | повертає кількість символів в `s` |\n",
    "| `char* strset(char* s, char c)` | заповнює `s` символами, код яких дорівнює `c`, повертає вказівник на `s` |\n",
    "| `char* strnset(char* s, char c, int n)` | замінює перші `n` символів `s` символами, код яких дорівнює `c`, повертає вказівник на `s` |\n",
    "\n",
    "## Функції для перевірки символів (стандартна бібліотека `ctype.h`)\n",
    "\n",
    "Для перевірки символів використовують функції, які повертають значення «істина» або «хиба». Прототипи цих функцій описано у файлі `ctype.h` бібліотеки стандартних функцій.\n",
    "\n",
    "Функції з бібліотеки повертають значення «істина», якщо:\n",
    "- `isalpha(c)`: `c` — символ алфавіту;\n",
    "- `isupper(c)`: `c` — символ верхнього регістру;\n",
    "- `islower(c)`: `c` — символ нижнього регістра;\n",
    "- `isdigit(c)`: `c` — цифра від 0 до 9;\n",
    "- `isxdigit(c)`: `c` — шістнадцяткова цифра;\n",
    "- `isalnum(c)`: `c` — буква чи цифра;\n",
    "- `isspace(c)`: `c` — символ пробілу, табуляції, переведення рядка чи формату."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```{margin}\n",
    "Запустити цей код можна за [посиланням](https://repl.it/@GasperPaul/Lab3-Example-34). \n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Приклад 3.4\n",
    "Наведена нижче програма демонструє використання функцій із бібліотеки `string.h`.\n",
    "\n",
    "```c\n",
    "#include <stdio.h>\n",
    "#include <string.h>\n",
    "\n",
    "int main()\n",
    "{\n",
    "    char m1[80] = \"Перший рядок\";\n",
    "    char m2[80] = \"другий рядок\";\n",
    "    char m3[80];\n",
    "\n",
    "    strncpy(m3, m1, 6); /* не додає '\\0' у кінці рядка */\n",
    "    printf(\"Результат strncpy(m3, m1, 6):\\n%s\\n\\n\", m3);\n",
    "\n",
    "    strcpy(m3, m1);\n",
    "    printf(\"Результат strcpy(m3, m1):\\n%s\\n\\n\", m3);\n",
    "\n",
    "    printf(\"Результат strcmp(m3, m1):\\n%d\\n\\n\", strcmp(m3, m1));\n",
    "  \n",
    "    strncat(m3, m2, 4);\n",
    "    printf(\"Результат strncat(m3, m2, 4):\\n%s\\n\\n\", m3);\n",
    "\n",
    "    strcat(m3, m2);\n",
    "    printf(\"Результат strcat(m3, m2):\\n%s\\n\\n\", m3);\n",
    "\n",
    "    printf(\"Кількість символів у рядку m1 дорівнює strlen(m1):\\n%lu\\n\\n\", strlen(m1));\n",
    "\n",
    "    memset(m3, 'f', 6);\n",
    "    printf(\"Результат strnset(m3, 'f', 6):\\n%s\\n\\n\", m3);\n",
    "}\n",
    "```"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.6.3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
